/* ex: set tabstop=2 expandtab: */
/* iriver-method.c: Gnome-VFS Iriver method

   Copyright (C) 2003 Michele Baldessari 
   
   The Gnome Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The Gnome Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the Gnome Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Authors: Michele Baldessari <michele@pupazzo.org> 
*/

#include <config.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <libgnomevfs/gnome-vfs.h>
#include <libgnomevfs/gnome-vfs-mime.h>

#include <libgnomevfs/gnome-vfs-method.h>
#include <libgnomevfs/gnome-vfs-module.h>
#include <libgnomevfs/gnome-vfs-module-callback-module-api.h>
#include <libgnomevfs/gnome-vfs-standard-callbacks.h>

#include <iriver.h>
#include "iriver-i18n.h"

#define DEBUG_IRIVER_ENABLE
#define DEBUG_IRIVER_LOCKS

#ifdef DEBUG_IRIVER_ENABLE
#define DEBUG_IRIVER(x)   g_print x
#else
#define DEBUG_IRIVER(x) 
#endif

#ifdef DEBUG_IRIVER_LOCKS
#define LOCK_IRIVER()     {g_mutex_lock (iriver_lock); g_print ("LOCK %s\n", G_GNUC_PRETTY_FUNCTION);}
#define UNLOCK_IRIVER()   {g_print ("UNLOCK %s\n", G_GNUC_PRETTY_FUNCTION); g_mutex_unlock (iriver_lock);}
#else
#define LOCK_IRIVER()     g_mutex_lock (iriver_lock)
#define UNLOCK_IRIVER()   g_mutex_unlock (iriver_lock)
#endif

#define IFP_TMP_FILE      "iriver_tmp_file.mp3"
#define IFP_LOG_FILE      "/tmp/iriver.log"

  
#define NUM_RESULT_STRINGS 41
static const char * const
result_strings[NUM_RESULT_STRINGS] = {
  "GNOME_VFS_OK",
  "GNOME_VFS_ERROR_NOT_FOUND",
  "GNOME_VFS_ERROR_GENERIC",
  "GNOME_VFS_ERROR_INTERNAL",
  "GNOME_VFS_ERROR_BAD_PARAMETERS",
  "GNOME_VFS_ERROR_NOT_SUPPORTED",
  "GNOME_VFS_ERROR_IO",
  "GNOME_VFS_ERROR_CORRUPTED_DATA",
  "GNOME_VFS_ERROR_WRONG_FORMAT",
  "GNOME_VFS_ERROR_BAD_FILE",
  "GNOME_VFS_ERROR_TOO_BIG",
  "GNOME_VFS_ERROR_NO_SPACE",
  "GNOME_VFS_ERROR_READ_ONLY",
  "GNOME_VFS_ERROR_INVALID_URI",
  "GNOME_VFS_ERROR_NOT_OPEN",
  "GNOME_VFS_ERROR_INVALID_OPEN_MODE",
  "GNOME_VFS_ERROR_ACCESS_DENIED",
  "GNOME_VFS_ERROR_TOO_MANY_OPEN_FILES",
  "GNOME_VFS_ERROR_EOF",
  "GNOME_VFS_ERROR_NOT_A_DIRECTORY",
  "GNOME_VFS_ERROR_IN_PROGRESS",
  "GNOME_VFS_ERROR_INTERRUPTED",
  "GNOME_VFS_ERROR_FILE_EXISTS",
  "GNOME_VFS_ERROR_LOOP",
  "GNOME_VFS_ERROR_NOT_PERMITTED",
  "GNOME_VFS_ERROR_IS_DIRECTORY",
  "GNOME_VFS_ERROR_NO_MEMORY",
  "GNOME_VFS_ERROR_HOST_NOT_FOUND",
  "GNOME_VFS_ERROR_INVALID_HOST_NAME",
  "GNOME_VFS_ERROR_HOST_HAS_NO_ADDRESS",
  "GNOME_VFS_ERROR_LOGIN_FAILED",
  "GNOME_VFS_ERROR_CANCELLED",
  "GNOME_VFS_ERROR_DIRECTORY_BUSY",
  "GNOME_VFS_ERROR_DIRECTORY_NOT_EMPTY",
  "GNOME_VFS_ERROR_TOO_MANY_LINKS",
  "GNOME_VFS_ERROR_READ_ONLY_FILE_SYSTEM",
  "GNOME_VFS_ERROR_NOT_SAME_FILE_SYSTEM",
  "GNOME_VFS_ERROR_NAME_TOO_LONG",
  "GNOME_VFS_ERROR_SERVICE_NOT_AVAILABLE",
  "GNOME_VFS_NUM_ERRORS"
};


static usb_dev_handle *dh = NULL;
static gboolean found_iriver = FALSE;
static GMutex *iriver_lock = NULL;

typedef struct {
  ifp_file_info *list;
  gchar *dirname;
  guint pos;
  gboolean list_type_is_folder;
} DirectoryHandle;

typedef struct {
  gboolean is_read;
  /* handle for the temp file */
  GnomeVFSHandle *handle;
  /* For writing */
  gchar *filename;
  gint folder;
} FileHandle;

/* Debug handler */
static void iriver_debug_handler(const gchar *string) 
{
  FILE *f;
  f = fopen(IFP_LOG_FILE, "a");
  if (f != NULL) {
    fprintf(stderr, "%s", string);
    fprintf(f, "%s", string);
    fclose(f);
  }
}


/* Module entry points. */
GnomeVFSMethod *vfs_module_init     (const char     *method_name,
                                     const char     *args);
void            vfs_module_shutdown (GnomeVFSMethod *method);



static const char *
gnome_vfs_uri_get_basename (const GnomeVFSURI *uri)
{
  char *p;

  g_return_val_if_fail (uri != NULL, NULL);

  if (uri->text == NULL) {
    return NULL;
  }

  p = strchr (gnome_vfs_uri_get_path(uri), GNOME_VFS_URI_PATH_CHR);
  if (p == NULL) {
    return NULL;
  }

  p++;
  if (*p == '\0') {
    return NULL;
  }

  return p;
}

static char *
unix_filename_to_dos (const char *filename)
{
  int len, i;
  char *dos_filename;
  
  dos_filename = g_strdup (filename);
  len = strlen (dos_filename);
  for (i = 0; i < len; i++)
    if (dos_filename[i] == '/')
      dos_filename[i] = '\\';
  
  return dos_filename;
}

static void
remove_double_slash(char *path) {
  char *pos = path;
  while (*path != '\0') {
    if (*path == '/' && *(path+1) == '/') {
      path++;
      continue;
    }
    *pos++ = *path++;
  }
  *pos = '\0';
}


static GnomeVFSResult
iriver_to_gnome_vfs_error(gint error)
{
  switch(error) {
    case IRIVER_GENERIC_USB_ERROR:
    case IRIVER_NO_IFP:
    case IRIVER_NO_USB_BUS:
    case IRIVER_NO_DEVICE:
		case IRIVER_LOCAL_FILE_OPEN_ERROR:
		case IRIVER_UPLOAD_ERROR:
		case IRIVER_CANNOT_SEND_FILENAME:
      return GNOME_VFS_ERROR_IO;
    case IRIVER_NO_FREE_SPACE:
      return GNOME_VFS_ERROR_NO_SPACE;
		case IRIVER_FILE_NOT_FOUND:
			return GNOME_VFS_ERROR_NOT_FOUND;
    case IRIVER_OK:
      return GNOME_VFS_OK;
  }
}   

static GnomeVFSResult
do_get_file_info (GnomeVFSMethod *method,
      GnomeVFSURI *uri,
      GnomeVFSFileInfo *file_info,
      GnomeVFSFileInfoOptions options,
      GnomeVFSContext *context)
{
  gchar *name, *tmp;
  unsigned char ctl[4];
  unsigned char ctl8[8];
  int res, size = 0;
  ifp_file_info *info;
  
  LOCK_IRIVER();
  /* Is this root? */
  if (gnome_vfs_uri_has_parent (uri) == 0) {
    file_info->name = g_strdup ("/");
    file_info->valid_fields = 
      GNOME_VFS_FILE_INFO_FIELDS_TYPE |
      GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;
    file_info->type = GNOME_VFS_FILE_TYPE_DIRECTORY;
    file_info->mime_type = g_strdup ("x-directory/normal");
    UNLOCK_IRIVER();
    DEBUG_IRIVER (("do_get_file_info() (root) %s\n returned %s\n",
        gnome_vfs_uri_to_string (uri, 0),
        gnome_vfs_result_to_string(GNOME_VFS_OK)));
    return GNOME_VFS_OK;
  }

  name = g_strdup(gnome_vfs_uri_get_basename(uri));
  remove_double_slash(name);
  if (g_str_has_suffix (name, GNOME_VFS_URI_PATH_STR) == TRUE) {
    g_free(name);
    name = g_strdup(gnome_vfs_uri_extract_dirname(uri));
  }

  name = gnome_vfs_unescape_string (unix_filename_to_dos(name), "");
  /* Using tentative prototype from libiriver */
  res = ifp_get_info (dh, name, &info);
  if (res != IRIVER_OK) {
    DEBUG_IRIVER (("do_get_file_info(): error returned by ifp_get_info %s\n", ifp_error_to_string(res) ));
		UNLOCK_IRIVER ();
    return iriver_to_gnome_vfs_error(res);
  }
  
  file_info->name = gnome_vfs_unescape_string( gnome_vfs_uri_extract_short_name (uri), "");
  file_info->valid_fields = 
      GNOME_VFS_FILE_INFO_FIELDS_TYPE |
      GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE |
      GNOME_VFS_FILE_INFO_FIELDS_SIZE;
  if (info->is_file) 
    file_info->type = GNOME_VFS_FILE_TYPE_REGULAR;
  else
    file_info->type = GNOME_VFS_FILE_TYPE_DIRECTORY;

  file_info->mime_type = g_strdup(info->mimetype);
  file_info->size = info->size;

  /* FIXME: free correctly the file_info struct */ 
  DEBUG_IRIVER (("do_get_file_info() base %s mime %s size %d returned %s\n", file_info->name,
    file_info->mime_type, file_info->size, gnome_vfs_result_to_string(GNOME_VFS_OK)));
  g_free(name);
  UNLOCK_IRIVER();
  return GNOME_VFS_OK;
}

static GnomeVFSResult
do_open (GnomeVFSMethod *method,
   GnomeVFSMethodHandle **method_handle,
   GnomeVFSURI *uri,
   GnomeVFSOpenMode mode,
   GnomeVFSContext *context)
{
  FileHandle *handle = NULL;
  GnomeVFSResult res;
  GnomeVFSFileInfo *info;
  gboolean is_read = FALSE;
  gchar *name, *local;

  DEBUG_IRIVER(("do_open() %s mode %d\n",
        gnome_vfs_uri_to_string (uri, 0), mode));

  local = g_strdup_printf ("%s/%s", g_get_tmp_dir(), IFP_TMP_FILE);
  if (g_file_test (local, G_FILE_TEST_EXISTS) == TRUE)
    return GNOME_VFS_ERROR_TOO_MANY_OPEN_FILES;

  if (mode & GNOME_VFS_OPEN_READ)
    is_read = TRUE;
  else if (mode & GNOME_VFS_OPEN_WRITE)
    is_read = FALSE;
  else
    return GNOME_VFS_ERROR_INVALID_OPEN_MODE;
  
  info = gnome_vfs_file_info_new ();
  res = do_get_file_info(method, uri, info, 0, context);
  LOCK_IRIVER();
  if ((res != GNOME_VFS_OK) && (is_read == TRUE)) {
    UNLOCK_IRIVER();
    g_free (info);
    g_free (local);
    DEBUG_IRIVER (("do_open(): file not found on read open, result %s\n", 
      gnome_vfs_result_to_string(res)));
    return res;
  }
  
  if ((res == GNOME_VFS_ERROR_NOT_FOUND) && (is_read == TRUE)) {
    UNLOCK_IRIVER();
    g_free (info);
    g_free (local);
    DEBUG_IRIVER (("do_open(): file not found on read open\n"));
    return GNOME_VFS_ERROR_NOT_FOUND;
  }

  if ((info->type == GNOME_VFS_FILE_TYPE_DIRECTORY) && (is_read == TRUE)) {
    UNLOCK_IRIVER();
    g_free (info);
    g_free (local);
    DEBUG_IRIVER(("do_open(): file is a directory on a read open\n"));
    return GNOME_VFS_ERROR_IS_DIRECTORY;
  }

	DEBUG_IRIVER (("do_open(): before is_read test %d\n", is_read));
  
  if (is_read == TRUE) {
    GnomeVFSHandle *local_handle;
    gint i;
    struct stat st;
    
    name = g_strdup(gnome_vfs_uri_get_path(uri));
    remove_double_slash(name);
    name = gnome_vfs_unescape_string (unix_filename_to_dos(name), "");
    
    DEBUG_IRIVER(("do_open(): opening read-only :%s\n", name));
    /* Copying file to /tmp */  
    i = ifp_download_file(dh, name, local);
    if (i != IRIVER_OK) {
      g_free (name);
      UNLOCK_IRIVER ();
      DEBUG_IRIVER (("do_open(): problems downloading from iriver ifp_download returned %s\n", 
          ifp_error_to_string (i) ));
      return iriver_to_gnome_vfs_error(i);
    }
    DEBUG_IRIVER (("do_open(): %s copied to %s\n", name, local));
    
    if (gnome_vfs_open (&local_handle, local, GNOME_VFS_OPEN_READ) !=
      GNOME_VFS_OK) {
      unlink(local);
      g_free (name);
      UNLOCK_IRIVER();
      DEBUG_IRIVER (("do_open(): error opening %s tmp file for new handle\n", local));
      return GNOME_VFS_ERROR_IO;
    }
    
    DEBUG_IRIVER (("do_open(): opened %s file for new handle\n", local));

    handle = g_new(FileHandle, 1);
    handle->is_read = is_read;
    handle->handle = local_handle;
    handle->filename = NULL;
    g_free (name);
  } else { /* Open for writing */
    GnomeVFSHandle *local_handle;
    gchar *basename, *local;

    DEBUG_IRIVER(("do_open(): opening write\n"));

    basename = gnome_vfs_unescape_string(gnome_vfs_uri_get_basename (uri), "");
    local = g_strdup_printf ("file://%s/%s", g_get_tmp_dir(), IFP_TMP_FILE);
  
    DEBUG_IRIVER (("do_open(): %s\n", local));
    if (gnome_vfs_create (&local_handle, local,
        GNOME_VFS_OPEN_WRITE, FALSE, 0600) != GNOME_VFS_OK) {
      UNLOCK_IRIVER();
      DEBUG_IRIVER (("do_open(): error opening the file for writing %s\n",
          local));
      g_free(local);
      g_free(basename);
      return GNOME_VFS_ERROR_IO;
    }

    /* Create the handle */
    handle = g_new(FileHandle, 1);
    handle->is_read = is_read;
    handle->handle = local_handle;
    handle->filename = g_strdup(basename);
    g_free(local);
    g_free(basename);
  }
  *method_handle = (GnomeVFSMethodHandle *)handle;
  
  return GNOME_VFS_OK;
}

static GnomeVFSResult
do_close (GnomeVFSMethod *method,
    GnomeVFSMethodHandle *method_handle,
    GnomeVFSContext *context)
{
  FileHandle *handle = (FileHandle *)method_handle;
  GnomeVFSResult res = GNOME_VFS_OK;
  gchar *tmpstr, *dest;

  tmpstr = g_strdup_printf("%s/%s", g_get_tmp_dir(), IFP_TMP_FILE);
  if (handle->is_read == TRUE) {
    gint tmp;
    DEBUG_IRIVER(("do_close() read (i.e. copy from iriver)\n"));

    tmp = unlink (tmpstr);
    DEBUG_IRIVER (("do_close(): unlinked temp file %s - %d \n", 
        tmpstr, tmp));

    if (tmp != 0)
      res = GNOME_VFS_ERROR_IO;
    else
      res = GNOME_VFS_OK;
  } else { /* Ending a write */
    gint i;
    FILE *f;
    struct stat stat;
    
    DEBUG_IRIVER(("do_close() write \n"));
/*    uri = gnome_vfs_uri_new (handle->filename);
    dest = gnome_vfs_uri_to_string (g_strdup(gnome_vfs_uri_get_basename(uri)); */

    dest = g_strdup_printf("/%s", handle->filename);
    DEBUG_IRIVER (("do_close(): uploading song %s to %s .. %s \n",
          tmpstr, handle->filename, dest));
    
    i = ifp_upload_file(dh, dest, tmpstr);
    switch(i) {
      case IRIVER_LOCAL_FILE_OPEN_ERROR:
        g_free(tmpstr);
        UNLOCK_IRIVER ();
        DEBUG_IRIVER (("do_close() write: error ifpupload returned [%s]\n", 
            ifp_error_to_string(i) ));
        return GNOME_VFS_ERROR_GENERIC;
      case IRIVER_NO_FREE_SPACE:
        g_free(tmpstr);
        UNLOCK_IRIVER();
        DEBUG_IRIVER (("do_close() write: error no space left [%s]\n", 
            ifp_error_to_string(i) ));
        return GNOME_VFS_ERROR_NO_SPACE;
      case IRIVER_OK:
        unlink (tmpstr);
        res = GNOME_VFS_OK;
        break;
    }

  }

  DEBUG_IRIVER (("do_close(): uploading result: %d\n", gnome_vfs_result_to_string(res)));
  gnome_vfs_close (handle->handle);
  g_free (handle->filename);
  g_free (handle);

  UNLOCK_IRIVER();
  g_free(tmpstr);

  return res;
}

static GnomeVFSResult
do_read (GnomeVFSMethod *method,
   GnomeVFSMethodHandle *method_handle,
   gpointer buffer,
   GnomeVFSFileSize num_bytes,
   GnomeVFSFileSize *bytes_read,
   GnomeVFSContext *context)
{
  FileHandle *handle = (FileHandle *)method_handle;

  DEBUG_IRIVER (("do_read() %Lu bytes\n", num_bytes));

  return gnome_vfs_read (handle->handle, buffer, num_bytes, bytes_read);
}

static GnomeVFSResult
do_write (GnomeVFSMethod *method,
    GnomeVFSMethodHandle *method_handle,
    gconstpointer buffer,
    GnomeVFSFileSize num_bytes,
    GnomeVFSFileSize *bytes_written,
    GnomeVFSContext *context)
{
  FileHandle *handle = (FileHandle *)method_handle;

  DEBUG_IRIVER (("do_write() - handle %p - name %s - num_bytes %ud \n", handle, handle->filename, num_bytes));
  return gnome_vfs_write (handle->handle, buffer, num_bytes,
      bytes_written);
}

static GnomeVFSResult
do_create (GnomeVFSMethod *method,
     GnomeVFSMethodHandle **method_handle,
     GnomeVFSURI *uri,
     GnomeVFSOpenMode mode,
     gboolean exclusive,
     guint perm,
     GnomeVFSContext *context)

{
  DEBUG_IRIVER (("do_create() %s mode %d\n",
        gnome_vfs_uri_to_string (uri, 0), mode));

  return do_open (method, method_handle, uri, GNOME_VFS_OPEN_WRITE,
      context);
}



static GnomeVFSResult
do_open_directory (GnomeVFSMethod *method,
       GnomeVFSMethodHandle **method_handle,
       GnomeVFSURI *uri,
       GnomeVFSFileInfoOptions options,
       GnomeVFSContext *context)
{
  DirectoryHandle *directory_handle;
  ifp_file_info *list = NULL, *info = NULL;
  int res;
  char *name;

  DEBUG_IRIVER(("do_open_directory() %s\n",
    gnome_vfs_uri_to_string (uri, 0)));
  
  name = g_strdup(gnome_vfs_uri_get_basename(uri));
  if (name == NULL) {
    name = (char *)malloc(3);
    strcpy(name, "/");
  }
  name = gnome_vfs_unescape_string (unix_filename_to_dos (name), "");
  DEBUG_IRIVER(("do_open_directory() base %s\n",
    name));
  LOCK_IRIVER();

  res = ifp_get_info (dh, name, &info);
  if (res != IRIVER_OK) {
    g_free (name);
    UNLOCK_IRIVER ();
    DEBUG_IRIVER (("do_open_directory (): ifp_get_info returned %s\n",
        ifp_error_to_string(res) ));
    return iriver_to_gnome_vfs_error(res);
  }

  if (info->is_file) {
    g_free(name);
    UNLOCK_IRIVER ();
    DEBUG_IRIVER (("do_open_directory (): trying to open a file\n" ));
    return GNOME_VFS_ERROR_NOT_A_DIRECTORY;
  }

  if (info->name != NULL)
    g_free(info->name);
  if (info->mimetype != NULL)
    g_free(info->mimetype);
  g_free(info);

  /* file exists and is a dir .. */ 
	if (ifp_empty_dir(dh, name) == TRUE) {
		info = NULL;
	} else {
		res = ifp_ls (dh, name, &info);
		if (res != IRIVER_OK) {
			g_free (name);
			UNLOCK_IRIVER ();
			DEBUG_IRIVER (("do_open_directory (): ifp_ls returned %s\n",
				  ifp_error_to_string(res) ));
			return iriver_to_gnome_vfs_error(res);
		}
	}

  directory_handle = g_new0 (DirectoryHandle, 1);
  directory_handle->dirname = g_strdup(name);
  directory_handle->pos = 0;
  directory_handle->list = info;

  *method_handle = (GnomeVFSMethodHandle *) directory_handle;

  g_free(name);
  UNLOCK_IRIVER();
  DEBUG_IRIVER(("do_open_directory end\n"));
  return GNOME_VFS_OK;
}


static GnomeVFSResult
do_close_directory (GnomeVFSMethod *method,
        GnomeVFSMethodHandle *method_handle,
        GnomeVFSContext *context)
{
  DirectoryHandle *directory_handle = (DirectoryHandle *) method_handle;
  DEBUG_IRIVER(("do_close_directory: start \n"));
	LOCK_IRIVER ();
	
	if (directory_handle == NULL)
    return GNOME_VFS_OK;
  if (directory_handle->dirname != NULL)
    g_free (directory_handle->dirname);

	DEBUG_IRIVER(("do_close_directory: before loop %p\n", 
			directory_handle->list));
  /* FIXME: need to free correctly all this */
  if (directory_handle != NULL) {
		ifp_free_file_list(&directory_handle->list);
    g_free (directory_handle);
  }
  DEBUG_IRIVER(("do_close_directory: end\n"));
	UNLOCK_IRIVER ();
  return GNOME_VFS_OK;
}


static GnomeVFSResult
do_read_directory (GnomeVFSMethod *method,
       GnomeVFSMethodHandle *method_handle,
       GnomeVFSFileInfo *file_info,
       GnomeVFSContext *context)
{
  DirectoryHandle *directory_handle = (DirectoryHandle *) method_handle;
  ifp_file_info *list;
  unsigned int i = 0;
  
	DEBUG_IRIVER (("do_read_directory() start\n" ));

  LOCK_IRIVER ();

  list = directory_handle->list; 
	if (list == NULL) {
		UNLOCK_IRIVER ();
		return GNOME_VFS_ERROR_EOF;
	}
  while (i < directory_handle->pos) {
    DEBUG_IRIVER (("do_read_directory() name % s i %d pos %d list %p list->next %p\n",
        list->name, i, directory_handle->pos, list, list->next));
    list = list->next;
    if (list == NULL) {
      UNLOCK_IRIVER();
      return GNOME_VFS_ERROR_EOF;
    }
    i++;
  }

	if (list == NULL) {
		UNLOCK_IRIVER ();
		return GNOME_VFS_ERROR_EOF;
	}

  GNOME_VFS_FILE_INFO_SET_LOCAL(file_info, FALSE);
  
  file_info->valid_fields = GNOME_VFS_FILE_INFO_FIELDS_NONE |
    GNOME_VFS_FILE_INFO_FIELDS_TYPE |
    GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE;

  if (list->is_file) {
    file_info->valid_fields |= GNOME_VFS_FILE_INFO_FIELDS_SIZE;
		file_info->type = GNOME_VFS_FILE_TYPE_REGULAR;
	}  else {
    file_info->type = GNOME_VFS_FILE_TYPE_DIRECTORY;
  } 
  
  file_info->size = list->size;
  file_info->name = g_strdup(list->name);
  file_info->mime_type = g_strdup(list->mimetype);

  directory_handle->pos++;
  UNLOCK_IRIVER();
  return GNOME_VFS_OK;
}


static gboolean
do_is_local (GnomeVFSMethod *method,
       const GnomeVFSURI *uri)
{
  return FALSE;
}

static GnomeVFSResult
do_make_directory (GnomeVFSMethod *method,
                  GnomeVFSURI *uri,
                  guint perm,
                  GnomeVFSContext *context)
{
  unsigned char ctl[4];
  gchar *name = NULL;
  GnomeVFSResult res;
  GnomeVFSFileInfo *info;
  int r;

  info = gnome_vfs_file_info_new ();
  res = do_get_file_info (method, uri, info, 0, context);
  
  DEBUG_IRIVER (("do_make_directory() starting removal of %s, get_file_info returned %s\n", 
      gnome_vfs_uri_get_basename(uri), gnome_vfs_result_to_string(res)));
  
  g_free(info);
  if (res == GNOME_VFS_OK) {
    return GNOME_VFS_ERROR_FILE_EXISTS;
  }

  name = g_strdup(gnome_vfs_uri_get_basename(uri));
  if (name == NULL) {
    name = (char *)malloc(3);
    strcpy(name, "/");
  }
  name = gnome_vfs_unescape_string (unix_filename_to_dos (name), "");
  DEBUG_IRIVER(("do_make_directory() base %s\n",
               name));

  LOCK_IRIVER ();


  r = ifp_make_dir (dh, name);
  if (r == IRIVER_OK) {
    g_free (name);
    UNLOCK_IRIVER ();
    return GNOME_VFS_OK;
  } else {
    DEBUG_IRIVER (("do_make_directory(): ifp_make_dir returned %s\n",
        ifp_error_to_string(r) ));
    g_free (name);
    UNLOCK_IRIVER ();
    /* FIXME: use saner error return codes ? */
    return GNOME_VFS_ERROR_IO;
  }
}

static GnomeVFSResult
do_remove_directory (GnomeVFSMethod *method,
      GnomeVFSURI *uri,
      GnomeVFSContext *context)
{
  gchar *name = NULL;
  unsigned char ctl[4];
  GnomeVFSResult res;
  GnomeVFSFileInfo *info;
  int i;
  
  
  info = gnome_vfs_file_info_new ();
  res = do_get_file_info(method, uri, info, 0, context);
  DEBUG_IRIVER (("do_remove_directory() starting removal of %s, get_file_info returned %s\n", 
      gnome_vfs_uri_get_basename(uri), gnome_vfs_result_to_string(res)));

  if (res == GNOME_VFS_ERROR_NOT_FOUND) 
    return GNOME_VFS_ERROR_NOT_FOUND;

  if (info->type == GNOME_VFS_FILE_TYPE_REGULAR)
    return GNOME_VFS_ERROR_NOT_A_DIRECTORY;
  
  LOCK_IRIVER();
  name = g_strdup(gnome_vfs_uri_get_basename(uri));
  DEBUG_IRIVER (("do_remove_directory() starting removal of %s\n", name));
  if (name == NULL) {
    name = (char *)malloc(3);
    strcpy(name, "/");
  }
  name = gnome_vfs_unescape_string (unix_filename_to_dos (name), "");

  /* Check if dir is empty */
  i = ifp_empty_dir (dh, name);
  if (i) {
    DEBUG_IRIVER (("Directory empty %d\n", i));
    /* FIXME: add error checking here */
    ifp_delete_dir(dh, name); 
    g_free(name);
    UNLOCK_IRIVER();
    return GNOME_VFS_OK;
  } else { /* dir is not empty */  
    UNLOCK_IRIVER();
    g_free(name);
    return GNOME_VFS_ERROR_DIRECTORY_NOT_EMPTY;
  }
}

static GnomeVFSResult
do_unlink (GnomeVFSMethod *method,
      GnomeVFSURI *uri,
      GnomeVFSContext *context)
{
  char *name;
  unsigned char ctl[4];
  GnomeVFSResult res;
  GnomeVFSFileInfo *info;

  DEBUG_IRIVER (("do_unlink() \n"));
  
  info = gnome_vfs_file_info_new ();
  res = do_get_file_info(method, uri, info, 0, context);

  if (res == GNOME_VFS_ERROR_NOT_FOUND) 
    return GNOME_VFS_ERROR_NOT_FOUND;
  
  
  LOCK_IRIVER();
  name = g_strdup(gnome_vfs_uri_get_basename(uri));
  DEBUG_IRIVER (("do_unlink() starting removal of %s\n", name));
  if (name == NULL) {
    name = (char *)malloc(3);
    strcpy(name, "/");
  }
  name = gnome_vfs_unescape_string (unix_filename_to_dos (name), "");
  
  if ((info->type == GNOME_VFS_FILE_TYPE_DIRECTORY)) {
    UNLOCK_IRIVER();
    g_free(name);
    return GNOME_VFS_ERROR_IS_DIRECTORY;
  } else {
    /* FIXME: add error checking here too */
    ifp_delete_file (dh, name);
    UNLOCK_IRIVER();
    g_free(name);
    return GNOME_VFS_OK;
  }
}

static GnomeVFSResult
do_check_same_fs (GnomeVFSMethod *method,
      GnomeVFSURI *a,
      GnomeVFSURI *b,
      gboolean *same_fs_return,
      GnomeVFSContext *context)
{
  DEBUG_IRIVER (("do_check_same_fs()\n"));
  *same_fs_return = FALSE;

  return GNOME_VFS_OK;
}

static GnomeVFSResult
do_move (GnomeVFSMethod *method,
   GnomeVFSURI *old_uri,
   GnomeVFSURI *new_uri,
   gboolean force_replace,
   GnomeVFSContext *context)
{
  GnomeVFSResult res;
  unsigned char ctl[4];
  gint i;
  gchar *new_name, *old_name;
  GnomeVFSFileInfo *old_info, *new_info;

  DEBUG_IRIVER (("do_move() %s %s\n",
        gnome_vfs_uri_to_string (old_uri, 0),
        gnome_vfs_uri_to_string (new_uri, 0)));

  old_info = gnome_vfs_file_info_new();
  res = do_get_file_info(method, old_uri, old_info, 0, context);
  
  DEBUG_IRIVER(("do_move(): lookup (src)\n"));
  
  /* if the original file isn't found, bail */
  if (res != GNOME_VFS_OK) {
    g_free (old_info);
    UNLOCK_IRIVER();
    return res;
  }

  new_info = gnome_vfs_file_info_new();
  res = do_get_file_info(method, new_uri, new_info, 0, context);
  DEBUG_IRIVER(("do_move(): lookup (dest) res %s\n", gnome_vfs_result_to_string(res)));

  /* if the dest file is found bail */
  if (res == GNOME_VFS_OK)
  {
    g_free (old_info);
    g_free (new_info);
    return GNOME_VFS_ERROR_NOT_SUPPORTED;
  }

  /* if the dest uri is invalid, bail */
  if (res == GNOME_VFS_ERROR_INVALID_URI)
  {
    g_free (old_info);
    g_free (new_info);
    return res;
  }

  /* if trying to move a dir to a dir (renaming included) bail out */
  if (old_info->type == GNOME_VFS_FILE_TYPE_DIRECTORY) {
    /* Check if src directory is empty, then operation is allowed, otherwise
     * not */

    LOCK_IRIVER();
    old_name = g_strdup(gnome_vfs_uri_get_basename(old_uri));
    DEBUG_IRIVER (("do_move() starting removal of %s\n", old_name));
    if (old_name == NULL) {
      g_free (old_info);
      g_free (new_info);
      UNLOCK_IRIVER ();
      return GNOME_VFS_ERROR_IO;
    }

    /* Check if dir is empty */
    i = ifp_empty_dir (dh, old_name);

    DEBUG_IRIVER (("do_move(): Directory %s empty %d\n", old_name, i));

    if (i == 1) { /* old_dir is not empty */  
      UNLOCK_IRIVER();
      g_free (old_name);
      g_free (old_info);
      g_free (new_info);
      return GNOME_VFS_ERROR_NOT_SUPPORTED;
    }
    /* old_dir is empty, so we simply remove it and create a new one with a new
     * name */
    /* FIXME: need to add a check that the two dirs are at the same depth
     * level */
    UNLOCK_IRIVER ();
    res = do_remove_directory (method, old_uri, context);
    if (res != GNOME_VFS_OK) {
      UNLOCK_IRIVER();
      g_free (old_name);
      g_free (old_info);
      g_free (new_info);
      return GNOME_VFS_ERROR_IO;
    }
    /* Create directory with new name */
    res = do_make_directory (method, new_uri, 0, context);  
    if (res != GNOME_VFS_OK) {
      UNLOCK_IRIVER();
      g_free (old_name);
      g_free (old_info);
      g_free (new_info);
      return GNOME_VFS_ERROR_IO;
    }
    g_free (old_info);
    g_free (new_info);
    g_free (old_name);
    return GNOME_VFS_OK;
  }

  g_free (old_info);
  g_free (new_info);
  return GNOME_VFS_ERROR_NOT_SUPPORTED;
}


static GnomeVFSMethod method = {
  sizeof (GnomeVFSMethod),
  do_open,              /* do_open */
  do_create,            /* do_create */
  do_close,             /* do_close */
  do_read,              /* do_read  */
  do_write,             /* do_write */
  NULL,                 /* do_seek */
  NULL,                 /* do_tell */
  NULL,                 /* do_truncate_handle */
  do_open_directory,    /* do_open_directory */
  do_close_directory,   /* do_close_directory */
  do_read_directory,    /* do_read_directory */
  do_get_file_info,     /* do_get_file_info */
  NULL,                 /* do_get_file_info_from_handle */
  do_is_local,          /* do_is_local */
  do_make_directory,    /* do_make_directory */
  do_remove_directory,  /* do_remove_directory */
  do_move,              /* do_move */
  do_unlink,            /* do_unlink */
  do_check_same_fs,     /* do_check_same_fs */
  NULL,                 /* do_set_file_info */
  NULL,                 /* do_truncate */
  NULL,                 /* do_find_directory */
  NULL                  /* do_create_symbolic_link */
};

GnomeVFSMethod *
vfs_module_init (const char *method_name, const char *args)
{
  int res;
  g_set_print_handler(iriver_debug_handler);

  res = ifp_init (&dh, IRIVER_INIT);
  if (res != IRIVER_OK) {
      DEBUG_IRIVER (("Initialization error %s\n", 
          ifp_error_to_string(res)));
      return NULL;
  }

  /* FIXME: Remove these once things look better ... */
  DEBUG_IRIVER (("Battery %d \n", 
      ifp_battery_condition(dh)));
  DEBUG_IRIVER (("Capacity %d bytes - Free %d bytes\n", 
      ifp_get_capacity(dh), ifp_get_free(dh)));
  
  iriver_lock = g_mutex_new();
  found_iriver = TRUE;

  return &method;
}

void
vfs_module_shutdown (GnomeVFSMethod *method)
{
  DEBUG_IRIVER (("Module unloaded!\n"));
  if (found_iriver && dh) {
    LOCK_IRIVER ();
    ifp_release (dh);
    UNLOCK_IRIVER ();
  }
  g_mutex_free(iriver_lock);
}
